from hwtypes import BitVector
from magma.array import Array
from magma.bit import Bit
from magma.bitutils import clog2, seq2int
from magma.circuit import coreir_port_mapping
from magma.generator import Generator2
from magma.interface import IO
from magma.protocol_type import MagmaProtocol
from magma.t import Type, In, Out
from magma.tuple import Product


class CoreIRCommonLibMuxN(Generator2):
    def __init__(self, N: int, width: int):
        self.name = f"coreir_commonlib_mux{N}x{width}"
        FlatT = Array[width, Bit]
        MuxInT = Product.from_fields("anon", dict(data=Array[N, FlatT],
                                                  sel=Array[clog2(N), Bit]))
        self.io = IO(I=In(MuxInT), O=Out(Array[width, Bit]))
        self.renamed_ports = coreir_port_mapping
        self.coreir_name = "muxn"
        self.coreir_lib = "commonlib"
        self.coreir_genargs = {"width": width, "N": N}
        self.primitive = True
        self.stateful = False

        def simulate(self, value_store, state_store):
            sel = BitVector[clog2(N)](value_store.get_value(self.I.sel))
            out = BitVector[width](value_store.get_value(self.I.data[int(sel)]))
            value_store.set_value(self.O, out)

        self.simulate = simulate


class Mux(Generator2):
    def __init__(self, height: int, T: Type):
        if issubclass(T, MagmaProtocol):
            T = T._to_magma_()
        T_str = str(T).replace("(", "").replace(")", "")\
                      .replace(",", "_").replace("=", "_")\
                      .replace("[", "").replace("]", "").replace(" ", "")
        self.name = f"Mux{height}x{T_str}"
        # TODO: Type must be hashable so we can cache.
        N = T.flat_length()

        io_dict = {}
        for i in range(height):
            io_dict[f"I{i}"] = In(T)
        if height == 2:
            io_dict["S"] = In(Bit)
        else:
            io_dict["S"] = In(Array[clog2(height), Bit])
        io_dict["O"] = Out(T)

        self.io = IO(**io_dict)
        mux = CoreIRCommonLibMuxN(height, N)()
        data = [Array[N, Bit](getattr(self.io, f"I{i}").flatten())
                for i in range(height)]
        mux.I.data @= Array[height, Array[N, Bit]](data)
        if height == 2:
            mux.I.sel[0] @= self.io.S
        else:
            mux.I.sel @= self.io.S
        self.io.O @= T.unflatten(mux.O.ts)


# NOTE(rsetaluri): We monkeypatch this function on to Array due to the circular
# dependency between Mux and Array. See the discussion on
# https://github.com/phanrahan/magma/pull/658.
def _dynamic_mux_select(this, key):
    return Mux(len(this), this.T)()(*this.ts, key)


Array.dynamic_mux_select = _dynamic_mux_select


def mux(I, S, **kwargs):
    if isinstance(S, Type) and S.const():
        S = seq2int(S.bits())
    if isinstance(S, int):
        return I[S]
    T = type(I[0])
    return Mux(len(I), T, **kwargs)()(*I, S)


# Monkey patch for ite impl without circular dependency
Bit._Mux = Mux



from functools import lru_cache
from magma.bit import Bit
from magma.bits import UInt
from magma.circuit import DeclareCoreirCircuit
from magma.t import In, Out


class BFloat(UInt):
    @classmethod
    @lru_cache(maxsize=None)
    def declare_unary_op(cls, op):
        N = len(cls)
        if N != 16:
            raise NotImplementedError("Only BFloat16 supported")
        coreir_genargs = {"exp_bits": 8, "frac_bits": 7}
        return DeclareCoreirCircuit(f"magma_BFloat_{N}_{op}",
                                    "I", In(cls),
                                    "O", Out(cls),
                                    coreir_name=op,
                                    coreir_genargs=coreir_genargs,
                                    coreir_lib="float")

    @classmethod
    @lru_cache(maxsize=None)
    def declare_binary_op(cls, op):
        N = len(cls)
        if N != 16:
            raise NotImplementedError("Only BFloat16 supported")
        coreir_genargs = {"exp_bits": 8, "frac_bits": 7}
        return DeclareCoreirCircuit(f"magma_BFloat_{N}_{op}",
                                    "I0", In(cls),
                                    "I1", In(cls),
                                    "O", Out(cls),
                                    coreir_name=op,
                                    coreir_genargs=coreir_genargs,
                                    coreir_lib="float")

    @classmethod
    @lru_cache(maxsize=None)
    def declare_compare_op(cls, op):
        N = len(cls)
        if N != 16:
            raise NotImplementedError("Only BFloat16 supported")
        coreir_genargs = {"exp_bits": 8, "frac_bits": 7}
        return DeclareCoreirCircuit(f"magma_BFloat_{N}_{op}",
                                    "I0", In(cls),
                                    "I1", In(cls),
                                    "O", Out(Bit),
                                    coreir_name=op,
                                    coreir_genargs=coreir_genargs,
                                    coreir_lib="float")

    @classmethod
    @lru_cache(maxsize=None)
    def declare_ite(cls, T):
        t_str = str(T)
        # Sanitize
        t_str = t_str.replace("(", "_")
        t_str = t_str.replace(")", "")
        t_str = t_str.replace("[", "_")
        t_str = t_str.replace("]", "")
        N = len(cls)
        if N != 16:
            raise NotImplementedError("Only BFloat16 supported")
        coreir_genargs = {"exp_bits": 8, "frac_bits": 7}
        return DeclareCoreirCircuit(f"magma_BFloat_{N}_ite_{t_str}",
                                    "I0", In(T),
                                    "I1", In(T),
                                    "S", In(Bit),
                                    "O", Out(T),
                                    coreir_name="mux",
                                    coreir_genargs=coreir_genargs,
                                    coreir_lib="float")



import os
import subprocess
from .coreir_ import InsertWrapCasts
from ..compiler import Compiler
from ..frontend import coreir_ as coreir_frontend
from ..is_definition import isdefinition
from ..passes import InstanceGraphPass
from ..passes.insert_coreir_wires import InsertCoreIRWires
from ..logging import root_logger
from magma.config import EnvConfig, config


_logger = root_logger()
config._register(
    fast_coreir_verilog_compile=EnvConfig(
        "MAGMA_FAST_COREIR_VERILOG_COMPILE", False),
)


def _make_verilog_cmd(deps, basename, opts):
    cmd = "coreir"
    if deps:
        cmd += f" -l {','.join(deps)}"
    cmd += f" -i {basename}.json"
    if opts.get("split", ""):
        split = opts["split"]
        cmd += f" -o \"{split}/*.v\" -s"
    else:
        cmd += f" -o {basename}.v"
    if opts.get("inline", False):
        cmd += " --inline"
    if opts.get("verilator_debug", False):
        cmd += " --verilator_debug"
    if opts.get("disable_width_cast", False):
        cmd += " --disable-width-cast"
    return cmd


def _make_opts(backend, opts):
    out = {}
    user_namespace = opts.get("user_namespace", None)
    if user_namespace is not None:
        out["user_namespace"] = backend.context.new_namespace(user_namespace)
    return out


class CoreIRCompiler(Compiler):
    def __init__(self, main, basename, opts):
        super().__init__(main, basename, opts)
        self.namespaces = self.opts.get("namespaces", ["global"])
        self.passes = opts.get("passes", [])
        if "markdirty" not in self.passes:
            self.passes.append("markdirty")
        self.deps = self._deps()
        self.backend = coreir_frontend.GetCoreIRBackend()

    def compile(self):
        InsertCoreIRWires(self.main).run()
        InsertWrapCasts(self.main).run()
        backend = self.backend
        opts = _make_opts(backend, self.opts)
        backend.compile(self.main, opts)
        backend.context.run_passes(self.passes, self.namespaces)
        output_json = (self.opts.get("output_intermediate", False) or
                       not self.opts.get("output_verilog", False) or
                       not config.fast_coreir_verilog_compile)
        if output_json:
            filename = f"{self.basename}.json"
            if isdefinition(self.main):
                backend.context.set_top(backend.modules[self.main.coreir_name])
            backend.context.save_to_file(filename, include_default_libs=False)
        if self.opts.get("output_verilog", False):
            fn = (self._fast_compile_verilog
                  if config.fast_coreir_verilog_compile
                  else self._compile_verilog)
            fn()
            self._compile_verilog_epilogue()
            return
        has_header_or_footer = (self.opts.get("header_file", "") or
                                self.opts.get("header_str", "") or
                                self.opts.get("footer_str", ""))
        if has_header_or_footer:
            _logger.warning("[coreir-compiler] header/footer only supported "
                            "when output_verilog=True, ignoring")

    def _compile_verilog(self):
        cmd = _make_verilog_cmd(self.deps, self.basename, self.opts)
        ret = subprocess.run(cmd, shell=True).returncode
        if ret:
            raise RuntimeError(f"CoreIR cmd '{cmd}' failed with code {ret}")

    def _fast_compile_verilog(self):
        top = self.backend.modules[self.main.coreir_name]
        filename = f"{self.basename}.v"
        opts = dict(
            libs=self.deps,
            split=self.opts.get("split", ""),
            inline=self.opts.get("inline", False),
            verilator_debug=self.opts.get("verilator_debug", False),
            disable_width_cast=self.opts.get("disable_width_cast", False),
        )
        ret = self.backend.context.compile_to_verilog(top, filename, **opts)
        if not ret:
            raise RuntimeError(f"CoreIR compilation to verilog failed")

    def _compile_verilog_epilogue(self):
        self._process_header_footer()

        bind_files = []
        # TODO(leonardt): We need fresh bind_files for each compile call.
        for name, file in self.backend.sv_bind_files.items():
            bind_files.append(f"{name}.sv")
            filename = os.path.join(os.path.dirname(self.basename), name)
            with open(f"{filename}.sv", "w") as f:
                f.write(file)

        if bind_files:
            bind_listings_file = self.basename + "_bind_files.list"
            with open(bind_listings_file, "w") as f:
                f.write("\n".join(bind_files))

        if self.opts.get("sv", False):
            if self.opts.get("split", False):
                _logger.warning("sv not supported with split, ignoring")
            else:
                args = (f"{self.basename}.v", f"{self.basename}.sv")
                subprocess.run(["mv", *args])

    def _deps(self):
        deps = self.opts.get("coreir_libs", set())
        pass_ = InstanceGraphPass(self.main)
        pass_.run()
        for key, _ in pass_.tsortedgraph:
            if key.coreir_lib:
                deps.add(key.coreir_lib)
            elif hasattr(key, "wrappedModule"):
                deps |= key.coreir_wrapped_modules_libs_used
        deps |= set(self.namespaces)
        return deps

    def _process_header_footer(self):
        header = ""
        if self.opts.get("header_file", ""):
            with open(self.opts["header_file"], "r") as header_file:
                header = header_file.read() + "\n"
        if self.opts.get("header_str", ""):
            header += self.opts["header_str"] + "\n"
        footer = self.opts.get("footer_str", "")
        if not header and not footer:
            return
        if self.opts.get("split", ""):
            _logger.warning("header/footer not supported with split, ignoring")
            return
        with open(f"{self.basename}.v", "r") as verilog_file:
            verilog = verilog_file.read()
        with open(f"{self.basename}.v", "w") as verilog_file:
            verilog_file.write(header + verilog + "\n" + footer)



from .utils import check_files_equal
from magma import compile

def compile_and_regress(callee_file, name, vendor, main, *largs, **kwargs):
    build = 'build/' + name
    gold = 'gold/' + name

    compile(build, main(*largs, **kwargs))
    assert check_files_equal(callee_file, build+'.v', gold+'.v')

    if vendor in ['altera', 'xilinx', 'lattice']:
        if vendor == 'altera' or vendor == 'xilinx':
            constraints = '.ucf'
        if vendor == 'lattice': 
            constraints = '.pcf'
        assert check_files_equal(callee_file,
            build+constraints, gold+constraints)



import colorlog
import inspect
import io
import logging
import sys
import traceback
from .backend.util import make_relative
from .config import config, EnvConfig


config._register(
    log_stream=EnvConfig("MAGMA_LOG_STREAM", "stderr"),
    log_level=EnvConfig("MAGMA_LOG_LEVEL", "INFO"),
    include_traceback=EnvConfig("MAGMA_INCLUDE_WIRE_TRACEBACK", False, bool),
    traceback_limit=EnvConfig("MAGMA_ERROR_TRACEBACK_LIMIT", 5, int),
)


def _make_bold(string):
    return f"\033[1m{string}\033[0m"


def _get_source_line(filename, lineno):
    with open(filename, "r") as f:
        return f.readlines()[lineno - 1]


def _attach_debug_info(msg, debug_info):
    file = debug_info[0]
    line = debug_info[1]
    line_info = _make_bold(f"{make_relative(file)}:{line}")
    msg = f"{line_info}: {msg}"
    try:
        source = _get_source_line(file, line).rstrip("\n")
        source = f">> {source}"
    except FileNotFoundError:
        source = f"(Could not file file {file})"
    msg = f"{msg}\n{source}"
    return msg


def _attach_traceback(msg, frame_selector, limit):
    """
    Attaches traceback string to @msg and returns new string.

    @frame_selector is a function which takes a list of stack frames and selects
    one. For example, it could select the frame based on an index, or based on
    the function names.
    """
    frame = frame_selector(inspect.stack()).frame
    with io.StringIO() as io_:
        traceback.print_stack(f=frame, limit=limit, file=io_)
        tb = io_.getvalue()
    msg = f"{msg}\n{tb}"
    return msg


def _frame_selector(frames):
    return frames[3]


def _get_additional_kwarg(kwargs, key):
    try:
        value = kwargs.pop(key)
        return value
    except KeyError:
        return None


class _MagmaLogger(logging.Logger):
    """
    Derivative of logging.Logger class, with two additional keyword args:
    * 'debug_info': Tuple of (file_name, line_no). If 'debug_info' is included,
       this source-level information is logged along with the message.
    * 'include_traceback': If True, a traceback is printed along with the
       message.
    """
    @staticmethod
    def __with_preamble(fn, msg, *args, **kwargs):
        debug_info = _get_additional_kwarg(kwargs, "debug_info")
        if debug_info:
            msg = _attach_debug_info(msg, debug_info)
        include_traceback = _get_additional_kwarg(kwargs, "include_traceback")
        if include_traceback or config.include_traceback:
            msg = _attach_traceback(
                msg, _frame_selector, config.traceback_limit)
        fn(msg, *args, **kwargs)

    def debug(self, msg, *args, **kwargs):
        _MagmaLogger.__with_preamble(super().debug, msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        _MagmaLogger.__with_preamble(super().info, msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        _MagmaLogger.__with_preamble(super().warning, msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        _MagmaLogger.__with_preamble(super().error, msg, *args, **kwargs)


# Set logging class to _MagmaLogger to override logging behavior. Also, setup
# root logger parameters.
logging.setLoggerClass(_MagmaLogger)
_log_stream = getattr(sys, config.log_stream)
_root_logger = logging.getLogger("magma")
_handler = colorlog.StreamHandler(_log_stream)
_handler.setFormatter(colorlog.ColoredFormatter(
    '%(log_color)s%(levelname)s%(reset)s:%(name)s:%(message)s'))
_root_logger.addHandler(_handler)
_root_logger.setLevel(config.log_level)


def root_logger():
    return logging.getLogger("magma")



from typing import Tuple

from hwtypes import BitVector

from magma.bit import Bit
from magma.bits import Bits
from magma.bitutils import clog2
from magma.braid import fork
from magma.circuit import coreir_port_mapping
from magma.conversions import from_bits, as_bits
from magma.interface import IO
from magma.t import Kind, In, Out
from magma.generator import Generator2


class _CoreIRLUT(Generator2):
    """
    Internally used generator for CoreIR LUT primitive
    """

    def __init__(self, contents: BitVector):
        n = clog2(len(contents))
        self.name = f"coreir_lut{str(int(contents))}"
        self.io = IO(I=In(Bits[n]), O=Out(Bit))
        self.renamed_ports = coreir_port_mapping

        def _simulate(self, value_store, state_store):
            in_ = BitVector[n](value_store.get_value(getattr(self, "in")))
            value_store.set_value(self.out, contents.as_bool_list()[in_])
        self.simulate = _simulate
        self.coreir_name = "lutN"
        self.coreir_lib = "commonlib"
        self.coreir_genargs = {"N": n}
        self.coreir_configargs = {"init": contents}


def _to_int(value):
    if isinstance(value, Bit):
        if not value.const():
            raise ValueError(f"Unexpected Bit value: {value}")
        return 1 if value is Bit.VCC else 0
    if not isinstance(value, Bits):
        value = as_bits(value)
    return int(value)


class LUT(Generator2):
    """
    Generate a LUT containing entries of a generic Type
    """

    def __init__(self, T: Kind, n: int, contents: Tuple):
        """
        T: The type of each LUT entry (e.g. Bit or Bits[5])

        n: nubmer of entries in the LUT

        contents: the contents of the LUT (a tuple of length `n` containing
                  elements of type T that are constant)
        """
        num_bits_per_entry = T.flat_length()
        self.io = IO(I=In(Bits[clog2(n)]), O=Out(T))

        contents = tuple(BitVector[num_bits_per_entry](_to_int(c))
                         for c in contents)
        # create LUT for each bit in T
        luts = []
        for i in range(num_bits_per_entry):
            bv = BitVector[n]([elem[i] for elem in contents])
            luts.append(_CoreIRLUT(bv)())
        self.io.O @= from_bits(T, fork(luts)(as_bits(self.io.I)))



from magma.bit import Bit
from magma.bits import Bits, UInt
from magma.bitutils import clog2
from magma.common import deprecated
from magma.t import Direction
from magma.conversions import array, zext, uint
from magma.protocol_type import MagmaProtocol


def get_slice(value: Bits, start: Bits, width: int):
    """
    Dynamic slice based off the value of `start` with constant `width`.

    Example:
    ```
    class TestSlice(m.Circuit):
        # Slice a 6 bit window of I using x as the start index
        IO = m.IO(
            I=m.In(m.Bits[10]),
            x=m.In(m.Bits[2]),
            O=m.Out(m.Bits[6])
        )

        io.O @= m.get_slice(io.I, start=io.x, width=6)
    ```

    Notes:
      * For now we only support slicing bits with bits index
      * We could add support for slicing Arrays
      * We could restrict index to be a UInt
    """
    # Construct an array where the index `i` is the slice of bits from `i` to
    # `i+width`, index into this array using `start`.
    return array([value[i:i + width] for i in range(len(value) - width)])[start]


@deprecated(msg="m.slice will not be supported in future versions, use "
            "m.get_slice instead")
def slice(*args, **kwargs):
    return get_slice(*args, **kwargs)


def set_slice(target: Bits, value: Bits, start: UInt, width: int):
    """
    target: the output with a dynamic slice range replaced by value
    value: the output driving a slice of target
    start: dynamic start index of the slice
    width: constant slice width
    """
    T = type(target)
    if issubclass(T, MagmaProtocol):
        T = T._to_magma_()
    output = T.qualify(Direction.Undirected)()
    orig_start_len = len(start)
    if len(start) < clog2(len(target)):
        start = zext(start, clog2(len(target)) - len(start))
    for i in range(len(target)):
        in_slice_range = Bit(True)
        # guards to avoid constant compare verilator error
        if i < (1 << orig_start_len) - 1:
            in_slice_range &= start <= i
        if i > 0:
            in_slice_range &= i <= (start + width - 1)
        output[i] @= in_slice_range.ite(value[uint(i, clog2(len(target))) -
                                              start], target[i])
    return output



from magma.simulator import PythonSimulator
from hwtypes import BitVector

class Coroutine:
    """
    Makes the initial call to __next__ upon construction to immediately
    start the routine.

    Overrides __getattr__ to support inspection of the local variables
    """
    def __init__(self, *args, **kwargs):
        self.reset(*args, **kwargs)

    @classmethod
    def definition(cls, *args, **kwargs):
        return cls._definition(*args, **kwargs)

    def reset(self, *args, **kwargs):
        self.co = self.definition(*args, **kwargs)
        next(self.co)

    def __getattr__(self, key):
        return self.co.gi_frame.f_locals[key]

    def send(self, args):
        return self.co.send(args)

    def __next__(self):
        return next(self.co)

    def next(self):
        return self.__next__()

def coroutine(func):
    class _Coroutine(Coroutine):
        _definition = func

    return _Coroutine


def check(circuit, sim, number_of_cycles, inputs_generator=None):
    simulator = PythonSimulator(circuit, clock=circuit.CLK)
    failed = False
    for cycle in range(number_of_cycles):
        if inputs_generator is None:
            next(sim)
        else:
            inputs = []
            for name, port in circuit.interface.ports.items():
                if name in ["CLK", "CE"]:
                    continue  # Skip clocks, TODO: Check the type
                if port.is_output():  # circuit input
                    input_value = getattr(inputs_generator, name)
                    inputs.append(input_value)
                    simulator.set_value(getattr(circuit, name), input_value)
            next(inputs_generator)
            if len(inputs) > 1:
                sim.send(inputs)
            elif len(inputs) == 1:
                sim.send(inputs[0])
            else:
                next(sim)
        simulator.advance(2)
        # Coroutine has an implicit __next__ call on construction so it already
        # is in it's initial state
        for name, port in circuit.interface.ports.items():
            if port.is_input():  # circuit output
                if getattr(sim, name) != BitVector(simulator.get_value(getattr(circuit, name))):
                    print(f"Failed on cycle {cycle}, port {name}, expected {getattr(sim, name)}, got {BitVector(simulator.get_value(getattr(circuit, name)))}")
                    failed = True
    assert not failed, "Failed to pass simulation"

def testvectors(circuit, sim, number_of_cycles, inputs_generator=None):
    outputs = []
    inputs = []
    for cycle in range(number_of_cycles):
        next_inputs = []
        in_ports = {}
        if inputs_generator is not None:
            for name, port in circuit.interface.ports.items():
                if name in ["CLK", "CE"]:
                    continue  # Skip clocks, TODO: Check the type
                if port.is_output():  # circuit input
                    input_value = getattr(inputs_generator, name)
                    next_inputs.append(input_value)
                    in_ports[name] = input_value
            next(inputs_generator)
        if len(next_inputs) > 1:
            sim.send(next_inputs)
        elif len(next_inputs) == 1:
            sim.send(next_inputs[0])
        else:
            next(sim)
        out_ports = {}
        for name, port in circuit.interface.ports.items():
            if name == "CLK":
                continue
            if port.is_input():  # circuit output
                out_ports[name] = getattr(sim, name)
        outputs.append(out_ports)
        inputs.append(in_ports)
    return inputs, outputs



from magma.primitives.lut import LUT
from magma.primitives.memory import Memory
from magma.primitives.mux import Mux, mux
from magma.primitives.register import Register
from magma.primitives.reduce import reduce
from magma.primitives.slice import get_slice, set_slice, slice
